;; term.line -- readline and other line level terminal utilities

(define-module term.line
  (use term.stty)
  (use term.screen)
  (use term.command)
  (use term.read-key)
  (use util.list)
  (use util.queue)
  (use util.history)
  (use srfi-2)
  (use srfi-13)
  (use text.format)
  (use gauche.parameter)
  (export readline read-password with-progress-meter gauche-complete))
(select-module term.line)

(define *default-history* (make-parameter (make-history)))

(define (string-common-prefix . l-o-s)
  (define (common-prefix a b)
    (with-string-io a
      (lambda ()
        (call-with-input-string b
          (lambda (p)
            (let loop ((ch1 (read-char))
                       (ch2 (read-char p)))
              (cond
                ((or (eof-object? ch1) (eof-object? ch2))
                 #f)
                ((eqv? ch1 ch2)
                 (display ch1)
                 (loop (read-char) (read-char p))))))))))
  (let loop ((prefix (car l-o-s))
             (ls (cdr l-o-s)))
    (if (null? ls)
      prefix
      (let ((str (car ls)))
        (if (string-prefix? prefix str)
          (loop prefix (cdr ls))
          (let ((res (common-prefix prefix str)))
            (and (not (equal? res "")) (loop res (cdr ls)))))))))

(define (gauche-complete word form buf pos)
  (define top (interaction-environment))
  (define res '())
  (define searched (make-hash-table 'eq?))
  (define (search mod)
    (unless (hash-table-exists? searched mod)
      (hash-table-put! searched mod #t)
      (for-each
       (lambda (s) (push! res s))
       (filter (cut string-prefix? word <>)
               (map symbol->string (hash-table-keys (module-table mod)))))))
  (search top)
  (for-each search (module-imports top))
  (for-each search (module-precedence-list top))
  (sort res))

(define (readline prompt . args)
  (define (get-coercer type)
    (cond
      ((memq type (list <object> <string>)) identity)
      ((eq? type <number>) string->number)
      ((eq? type <symbol>) string->symbol)
      ((eq? type <char>)   (cut string-ref <> 0))
      (else read-from-string)))
  (define (get-predicate type)
    (cond
      ((memq type (list <object> <string>)) (lambda _ #t))
      ((eq? type <number>) string->number)
      ((eq? type <symbol>) (lambda _ #t)) ;; require non-null symbols?
      ((eq? type <char>)   (lambda (s) (= (string-length s) 1)))
      (else (lambda _ #t))))
  (let-keywords* args ((history (*default-history*))
                       (complete #f)
                       (highlight '())
                       (type <object>)
                       (default #f)
                       (coercer (get-coercer type))
                       (predicate (get-predicate type)))
    (with-read-key-io
     (lambda ()
       (let* ((buf "")
              (clipboard "")
              (pos 0)
              (done? #f)
              (cur (get-cursor))
              (dim (get-terminal-size))
              (y-orig (car cur))  ;; origin, where the prompt starts
              (x-orig (cdr cur))  ;; 
              (y-max (car dim))   ;; max screen dimensions
              (x-max (cadr dim))) ;;
         (define (init)
           (set! cur (get-cursor))
           (set! y-orig (car cur))
           (set! x-orig (cdr cur)))
         (define (word-before)
           (let* ((str (substring buf 0 pos))
                  (m (#/[-+.\/$%&~=\w]+\s*$/ str)))
             (and m (rxmatch-start m 0))))
         (define (word-after)
           (let* ((str (substring buf pos (string-length buf)))
                  (m (#/^\s*[-+.\/$%&~=\w]+/ str)))
             (and m (rxmatch-end m 0))))
         (define (parse)
           (let* ((str (substring buf 0 pos))
                  (m (#/[-+.\/$%&~=\w]+$/ str)))
             (values (if m (m 0) "") #f)))
         (define (insert str)
           (set! buf (string-append
                      (substring buf 0 pos)
                      str
                      (substring buf pos (string-length buf))))
           (inc! pos (string-length str))
           (if (= pos (string-length buf))
             (display str)
             (refresh)))
         (define (show)
           (display prompt)
           (let* ((cur2 (get-cursor))
                  (y (car cur2))        ;; loop variable
                  (x (cdr cur2))        ;;
                  (y-cur y)             ;; cursor, not current
                  (x-cur x))            ;;
             (let ((i 0))
               (for-each
                (lambda (l)
                  (unless (= y y-orig)
                    (set! x 1)
                    (newline))
                  (when (and (<= i pos) (>= (+ i (string-length l)) pos))
                    (set! y-cur y)
                    (set! x-cur (+ x (- pos i)))
                    (while (> x-cur x-max)
                      (inc! y-cur)
                      (dec! x-cur x-max)))
                  (inc! i (+ 1 (string-length l)))
                  (display l)
                  (inc! y)
                  (when (and (> y y-max) (not (= y-orig 1)))
                    (dec! y-orig)
                    (dec! y))
                  (display (term-code 'erase-in-line)))
                (string-split buf #/(?:\r?\n|\r)/)))
             ;; reset point
             (move-cursor y-cur x-cur)))
         (define (refresh)
           (move-cursor y-orig x-orig)
           (show))
         (history 'push buf)
         (display prompt)
         (let loop ((ch (read-key)))
           (case ch
             ((KEY-NEWLINE)
              (if (predicate buf)
                (set! done? #t)
                (insert "\n")))
             ((KEY-CONTROL-a)
              (set! pos 0)
              (refresh))
             ((KEY-CONTROL-d)
              (when (equal? buf "")
                (newline)
                (exit)))
             ((KEY-CONTROL-e)
              (set! pos (string-length buf))
              (refresh))
             ((KEY-CONTROL-w)
              (and-let* ((start (word-before)))
                (set! clipboard (substring buf start pos))
                (set! buf (string-append (substring buf 0 start)
                                         (substring buf pos (string-length buf))))
                (set! pos start)
                (refresh)))
             ((KEY-CONTROL-u)
              (set! clipboard (substring buf 0 pos))
              (set! buf (substring buf pos (string-length buf)))
              (set! pos 0)
              (refresh))
             ((KEY-CONTROL-k)
              (unless (or (zero? pos) (= 1 (car (get-cursor))))
                (dec! pos))
              (set! clipboard (substring buf pos (string-length buf)))
              (set! buf (substring buf 0 pos))
              (refresh))
             ((KEY-CONTROL-l)
              (set! x-orig 1)
              (set! y-orig 1)
              (clear-screen)
              (init)
              (refresh))
             ((KEY-LEFT KEY-CONTROL-b)
              (unless (zero? pos)
                (dec! pos)
                (if (= 1 (cdr (get-cursor)))
                  (refresh)
                  (display (term-code 'cursor-backward)))))
             ((KEY-RIGHT KEY-CONTROL-f)
              (unless (= pos (string-length buf))
                (inc! pos)
                (if (= x-max (cdr (get-cursor)))
                  (refresh)
                  (display (term-code 'cursor-forward)))))
             ((KEY-CONTROL-t)
              (when (and (not (zero? pos)) (>= (string-length buf) 2))
                (let ((len (string-length buf)))
                  (cond
                    ((not (= pos len))
                     (set! buf (string-append (substring buf 0 (- pos 1))
                                              (substring buf pos (+ pos 1))
                                              (substring buf (- pos 1) pos)
                                              (substring buf (+ pos 1) len)))
                     (inc! pos))
                    ((> pos 1)
                     (set! buf (string-append (substring buf 0 (- pos 2))
                                              (substring buf (- pos 1) pos)
                                              (substring buf (- pos 2) (- pos 1)))))))
                (refresh)))
             ((KEY-CONTROL-y)
              (insert clipboard))
             ;; handle aliased keys.... should think of a better
             ;; read-key API, but it's handy to be able to put
             ;; everything in a case statement like this
             ((KEY-SPACE)         (insert " "))
             ((KEY-EXCLAIM)       (insert "!"))
             ((KEY-QUOTEDBL)      (insert "\""))
             ((KEY-HASH)          (insert "#"))
             ((KEY-DOLLAR)        (insert "$"))
             ((KEY-AMPERSAND)     (insert "&"))
             ((KEY-QUOTE)         (insert "'"))
             ((KEY-LEFTPAREN)     (insert "("))
             ((KEY-RIGHTPAREN)    (insert ")"))
             ((KEY-ASTERISK)      (insert "*"))
             ((KEY-PLUS)          (insert "+"))
             ((KEY-MINUS)         (insert "-"))
             ((KEY-PERIOD)        (insert "."))
             ((KEY-SLASH)         (insert "/"))
             ((KEY-COLON)         (insert ":"))
             ((KEY-SEMICOLON)     (insert ";"))
             ((KEY-LESS)          (insert "<"))
             ((KEY-EQUALS)        (insert "="))
             ((KEY-GREATER)       (insert ">"))
             ((KEY-QUESTION)      (insert "?"))
             ((KEY-AT)            (insert "@"))
             ((KEY-LEFTBRACKET)   (insert "["))
             ((KEY-BACKSLASH)     (insert "\\"))
             ((KEY-RIGHTBRACKET)  (insert "]"))
             ((KEY-CARET)         (insert "^"))
             ((KEY-UNDERSCORE)    (insert "-"))
             ((KEY-BACKQUOTE)     (insert "`"))
             ((KEY-BACKSPACE)
              (unless (= pos 0)
                (set! buf (substring buf 0 (- (string-length buf) 1)))
                (dec! pos)
                (cond
                  ((= 1 (cdr (get-cursor)))
                   (display (term-code 'delete-line))
                   (refresh))
                  (else
                   (display (term-code 'cursor-backward))
                   (display #\space)
                   (display (term-code 'cursor-backward))))))
             ;; history
             ((KEY-UP KEY-CONTROL-p)
              (unless (history 'empty?)
                (history 'set buf)
                (set! buf (history 'back))
                (set! pos (string-length buf))
                (refresh)))
             ((KEY-DOWN KEY-CONTROL-n)
              (unless (history 'empty-forward?)
                (history 'set buf)
                (set! buf (history 'forward))
                (set! pos (string-length buf))
                (refresh)))
             ;; tab completion
             ((KEY-TAB KEY-CONTROL-i)
              (receive (word form) (parse)
                (cond
                  ((procedure? complete)
                   (and-let* ((res (complete word form buf pos)))
                     (cond
                       ((= (length res) 1)
                        (insert (substring (car res)
                                           (string-length word)
                                           (string-length (car res)))))
                       ((apply string-common-prefix
                               (map (lambda (s)
                                      (substring s (string-length word)
                                                 (string-length s)))
                                    res))
                        => insert)
                       (else
                        (newline)
                        (format-columns res (current-output-port) x-max)
                        (init)
                        (show))))))))
             ;; default, self insert
             (else
              (insert (key->string ch))))
           ;; loop or quit
           (cond
             (done?
              (newline)
              (history 'set buf)
              (history 'delete-forward "")
              (coercer buf))
             (else
              (loop (read-key))))))))))

(define (read-password prompt . args)
  (let-keywords* args ((echo-char #f))
    (with-read-key-io
      (lambda ()
        (display prompt)
        (begin0
          (if echo-char
            (let loop ((ch (read-char))
                       (res '()))
              (case ch
                ((#\x08 #\delete)
                 (cond
                   ((pair? res)
                    (display (term-code 'cursor-backward))
                    (display (term-code 'delete-character))
                    (loop (read-char) (cdr res)))
                   (else (loop (read-char) res))))
                ((#\x15)
                 (display (term-code 'delete-line))
                 (display #\return)
                 (display prompt)
                 (loop (read-char) '()))
                ((#\newline #\return)
                 (apply string (reverse res)))
                (else
                 (display (string echo-char) (current-output-port))
                 (loop (read-char) (cons ch res)))))
            (read-line))
          (newline))))))

;; like normal format, takes extra parameters and add extra specifiers:
;;  P - percent complete
;;  E - elapsed time in seconds
;;  R - estimated remaining time in seconds
;;  = - completion bar
(define (progress-format fmt pct start filler width . args)
  (let ((args2 '()))
    (let ((fmt2
           (regexp-replace-all #/~((?:[\d,]+|'.)*)(@?:?)(.)/ fmt
             (lambda (m)
               (string-append
                "~" (m 1) (m 2)
                (let ((spec (string-ref (m 3) 0)))
                  (case spec
                    ((#\p #\P)
                     (push! args2 (inexact->exact (round (* 100 pct)))) "D")
                    ((#\e #\E)
                     (push! args2 (- (sys-time) start)) "D")
                    ((#\r #\R)
                     (push! args2 (* (- 1.0 pct) (- (sys-time) start))) "D")
                    ((#\=)
                     (let ((len (inexact->exact (round (* pct width)))))
                       (push! args2 (string-append
                                     (make-string (max len 0) filler)
                                     (make-string (max (- width len) 0)
                                                  #\space))))
                     "A")
                    (else
                     (push! args2 (pop! args)) (m 3)))))))))
      (apply format fmt2 (reverse args2)))))

(define (with-progress-meter step percent . args)
  (let-keywords* args ((format "~3p%[~=]")
                       (filler #\=))
    (let ((start (sys-time))
          (width 40))
      (let loop ((res (step)))
        (let ((pct (percent res)))
          (display (term-code 'delete-line))
          (display #\return)
          (display (progress-format format pct start filler width))
          (flush)
          (if (>= pct 1.0)
            (begin (newline) res)
            (loop (step))))))))

(provide "term/line")
