;;;; read-key.scm -- abstract keyboard input as symbols

(define-module term.read-key
  (use term.stty)
  (export read-key key-ready? read-key-initialize read-key-restore
          key->string with-read-key-io))
(select-module term.read-key)

(define (read-key-initialize)
  ;;(sys-system "stty -icanon -echo")
  (stty '(not icanon echo)))

(define (read-key-restore)
  ;;(sys-system "stty sane")
  (stty 'sane))

(define (with-read-key-io thunk)
  (let* ((out (current-output-port))
         (buffering (port-buffering out)))
    (with-stty '(not icanon echo)
      (lambda ()
        (dynamic-wind
            (lambda () (set! (port-buffering out) :none))
            thunk
            (lambda () (set! (port-buffering out) buffering)))))))

(define (key-ready?)
  (char-ready? (current-input-port)))

(define (read-numeric-sequence)
  ;;(format #t "read-numeric-sequence\n")
  (let loop ((c (peek-char)) (acc '()))
    (case c
      ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
       (read-char) ;; skip peeked char
       (loop (peek-char) (cons c acc)))
      (else
       (string->number (apply string (reverse acc)))))))

(define escape-tilde-keys
  #(KEY-DUMMY-0 KEY-HOME KEY-INSERT KEY-DELETE KEY-END KEY-PAGEUP KEY-PAGEDOWN
    KEY-DUMMY-7 KEY-DUMMY-8 KEY-DUMMY-9 KEY-DUMMY-10 ;; unused, i think
    KEY-F1 KEY-F2 KEY-F3 KEY-F4 KEY-F5 KEY-DUMMY-6 KEY-F6 KEY-F7 KEY-F8 KEY-F9
    KEY-F10 KEY-DUMMY-11 KEY-F11 KEY-F12))

(define (read-escape-bracket-sequence)
  ;;(format #t "read-escape-bracket-sequence\n")
  (let ((c (peek-char)))
    (case c
      ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
       ;; numeric argument
       (let* ((arg (read-numeric-sequence))
              (key (read-char)))
         (case key
           ((#\~)
            (vector-ref escape-tilde-keys arg))
           (else
            (string->symbol (string-append "KEY-" (make-string 1 key)))))))
      ((#\A) (read-char) 'KEY-UP)
      ((#\B) (read-char) 'KEY-DOWN)
      ((#\C) (read-char) 'KEY-RIGHT)
      ((#\D) (read-char) 'KEY-LEFT)
      (else
       'KEY-CONTROL-SEQUENCE-INTRO))))

(define (read-escape-sequence)
  ;;(format #t "read-escape-sequence\n")
  (if (key-ready?)
    (let ((c (peek-char)))
      (case c
        ((#\[)
         (read-char) ;; read peeked
         (or (read-escape-bracket-sequence)
             'KEY-ESCAPE))
        ((#\newline)
         ;; fake a command key
         (read-char) ;; read peeked
         'KEY-COMMAND)
        (else
         ;; unknown command, send back escape with peeked char unread
         'KEY-ESCAPE)))
    ;; no char ready means literal escape
    'KEY-ESCAPE))

(define (read-key)
  ;;(format #t "read-key\n")
  (let ((c (read-char)))
    (case c

      ;; escapes
      ((#\escape)
       (read-escape-sequence))

      ;; controls (uppercase ASCII - 64)
      ((#\x01)       'KEY-CONTROL-a)
      ((#\x02)       'KEY-CONTROL-b)
      ((#\x03)       'KEY-CONTROL-c)
      ((#\x04)       'KEY-CONTROL-d)
      ((#\x05)       'KEY-CONTROL-e)
      ((#\x06)       'KEY-CONTROL-f)
      ((#\x07)       'KEY-CONTROL-g)
      ((#\x08)       'KEY-CONTROL-h)
      ((#\x09)       'KEY-TAB)     ;; CONTROL-i
      ((#\tab)       'KEY-TAB)     ;; CONTROL-i
      ((#\x0a)       'KEY-NEWLINE) ;; CONTROL-j
      ((#\newline)   'KEY-NEWLINE) ;; CONTROL-j
      ((#\x0b)       'KEY-CONTROL-k)
      ((#\x0c)       'KEY-CONTROL-l)
      ((#\x0d)       'KEY-RETURN)  ;; CONTROL-m
      ((#\return)    'KEY-RETURN)  ;; CONTROL-m
      ((#\x0e)       'KEY-CONTROL-n)
      ((#\x0f)       'KEY-CONTROL-o)
      ((#\x10)       'KEY-CONTROL-p)
      ((#\x11)       'KEY-CONTROL-q)
      ((#\x12)       'KEY-CONTROL-r)
      ((#\x13)       'KEY-CONTROL-s)
      ((#\x14)       'KEY-CONTROL-t)
      ((#\x15)       'KEY-CONTROL-u)
      ((#\x16)       'KEY-CONTROL-v)
      ((#\x17)       'KEY-CONTROL-w)
      ((#\x18)       'KEY-CONTROL-x)
      ((#\x19)       'KEY-CONTROL-y)
      ((#\x1a)       'KEY-CONTROL-z)

      ;; metas (160 + uppercase ASCII)
      ((#\xe1)       'KEY-META-a)
      ((#\xe2)       'KEY-META-b)
      ((#\xe3)       'KEY-META-c)
      ((#\xe4)       'KEY-META-d)
      ((#\xe5)       'KEY-META-e)
      ((#\xe6)       'KEY-META-f)
      ((#\xe7)       'KEY-META-g)
      ((#\xe8)       'KEY-META-h)
      ((#\xe9)       'KEY-META-i)
      ((#\xea)       'KEY-META-j)
      ((#\xeb)       'KEY-META-k)
      ((#\xec)       'KEY-META-l)
      ((#\xed)       'KEY-META-m)
      ((#\xee)       'KEY-META-n)
      ((#\xef)       'KEY-META-o)
      ((#\xf0)       'KEY-META-p)
      ((#\xf1)       'KEY-META-q)
      ((#\xf2)       'KEY-META-r)
      ((#\xf3)       'KEY-META-s)
      ((#\xf4)       'KEY-META-t)
      ((#\xf5)       'KEY-META-u)
      ((#\xf6)       'KEY-META-v)
      ((#\xf7)       'KEY-META-w)
      ((#\xf8)       'KEY-META-x)
      ((#\xf9)       'KEY-META-y)
      ((#\xfa)       'KEY-META-z)

      ;; mnemonics (meant to be compatible with SDL keynames)
      ((#\space)     'KEY-SPACE)
      ((#\!)         'KEY-EXCLAIM)
      ((#\")         'KEY-QUOTEDBL)
      ((#\#)         'KEY-HASH)
      ((#\$)         'KEY-DOLLAR)
      ((#\&)         'KEY-AMPERSAND)
      ((#\')         'KEY-QUOTE)
      ((#\()         'KEY-LEFTPAREN)
      ((#\))         'KEY-RIGHTPAREN)
      ((#\*)         'KEY-ASTERISK)
      ((#\+)         'KEY-PLUS)
      ((#\-)         'KEY-MINUS)
      ((#\.)         'KEY-PERIOD)
      ((#\/)         'KEY-SLASH)
      ((#\:)         'KEY-COLON)
      ((#\;)         'KEY-SEMICOLON)
      ((#\<)         'KEY-LESS)
      ((#\=)         'KEY-EQUALS)
      ((#\>)         'KEY-GREATER)
      ((#\?)         'KEY-QUESTION)
      ((#\@)         'KEY-AT)
      ((#\[)         'KEY-LEFTBRACKET)
      ((#\\)         'KEY-BACKSLASH)
      ((#\])         'KEY-RIGHTBRACKET)
      ((#\^)         'KEY-CARET)
      ((#\-)         'KEY-UNDERSCORE)
      ((#\`)         'KEY-BACKQUOTE)
      ((#\delete)    'KEY-BACKSPACE)

      ;; auto
      (else
       (string->symbol (string-append "KEY-" (make-string 1 c)))))))

(define (key->string key)
  (let ((str (symbol->string key)))
    (substring str 4 (string-length str))))

(provide "term/read-key")

