;;;; screen.scm -- curses-like commands for Scheme

;;; Created:    <2002-07-01 15:07:17 foof>
;;; Time-stamp: <2003-08-30 19:46:44 foof>
;;; Author:     Alex Shinn <foof@synthcode.com>


(define-module term.screen
  (use term.command)
  (use term.read-key)
  (use gauche.array)
  (use srfi-2)
  (use srfi-13)
  (export clear-screen move-cursor get-cursor get-terminal-size
          get-terminal-dim draw-screen string->screen string-width)
  )
(select-module term.screen)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic wrappers to vt100 commands

(define home-position (term-code 'cursor-position))

(define (clear-screen)
  (display (term-code 'erase-in-display 2))
  (display home-position)
  (flush))

(define (move-cursor y x)
  (display (term-code 'cursor-position y x)))

(define (cursor-back count . opt-width)
  (let* ((here (get-cursor))
         (y (car here))
         (x (+ count (cdr here)))
         (width (get-optional opt-width (get-terminal-width))))
    (while (>= x width)
      (set! x (- x width))
      (dec! y))
    (move-cursor y x)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; getting information about the screen

(define (add-digit n digit)
  (+ (* n 10) digit))

(define (skip-until c)
  (cond ((eq? (read-char) c)
         #t)
        ((eof-object? c)
         #f)
        (else
         (skip-until c))))

;; Adapted from post by Julian Fondren on gauche-devel ml 2002/08/21.
;; Note: return value is reversed to correspond with terminal y-major
;; perspective.  Could be made to use reverse parse of
;; cursor-position-report.

(define (get-cursor)
  (define (fail . args)
    (apply error args)) ; or return #f?
  (display (term-code 'device-status-report 6))
  (flush)
  (unless (skip-until #\escape)
    (fail "expected ESCAPE"))
  (unless (eq? (read-char) #\[)
    (fail "expected ["))
  (let next-ydigit ((c (read-char))
                    (y 0))
    (cond ((char-numeric? c)
           (next-ydigit (read-char) (add-digit y (digit->integer c))))
          ((eq? c #\;)
           (let next-xdigit ((c (read-char))
                             (x 0))
             (cond ((char-numeric? c)
                    (next-xdigit (read-char) (add-digit x (digit->integer c))))
                   ((eq? c #\R)
                    (cons y x))
                   (else
                    (fail "expected R")))))
          (else
           (fail "bad device-status-report char: " c)))))

(define terminal-dimension-limit 2048) ; just to be safe

(define (get-terminal-size)
  (with-read-key-io
   (lambda ()
     (let ((here (get-cursor)))
       (dynamic-wind
           (lambda () #t)
           (lambda ()
             (move-cursor terminal-dimension-limit
                          terminal-dimension-limit)
             (let1 x (get-cursor) (list (car x) (cdr x))))
           (lambda () (move-cursor (car here) (cdr here))))))))

(define get-terminal-height (compose car get-terminal-size))
(define get-terminal-width (compose cadr get-terminal-size))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; tools for managing screens as 2-dimensional arrays

(define screen-ref array-ref)
(define screen-set! array-set!)

;;; screen is a 2x2 array of chars
(define (draw-screen screen . y&x-off)
  (let ((y-off 0)
        (x-off 0))
    (cond ((not (null? y&x-off))
           (set! y-off (car y&x-off))
           (let ((tail (cdr y&x-off)))
             (if (not (null? tail))
               (set! x-off (car tail))))))
    ;;(format #t "draw-screen: ~S, y: ~S, x: ~S\n" screen y-off x-off)
    (array-for-each-index
     screen
     (lambda (y x)
       (if (= x 0)
         (move-cursor (+ 1 y y-off) (+ 1 x-off)))
       (let* ((char (screen-ref screen y x)))
         (display char))
       ))))

;; Currently (naively) assumes all ASCII chars are 1 wide, and other
;; chars are 2 wide.  Useful on Kanji terminals, which display Kanji as
;; double-width characters.
(define (char-width c)
  (if (< (char->integer c) 256) 1 2))

(define (string-width str)
  (string-fold (lambda (c sum) (+ (char-width c) sum)) 0 str))

(define (string->row str cols)
  (let ((res (string-fold
              (lambda (c cdr)
                (let ((diff (- (char-width c) 1)))
                  (if (= diff 0)
                    (cons c cdr)
                    (append (list c) (make-list diff #f) cdr))))
              '()
              str)))
    (let ((diff (- cols (length res))))
      ;; pad row to cols wide
      (if (<= diff 0)
        (reverse res)
        (append (reverse res) (make-list diff #\space))))))

(define (string->screen str)
  (let* ((lines (string-split str #\newline))
         (rows (length lines))
         (cols (apply max (map string-width lines))))
    ;;(format #t "lines: ~S\n" lines)
    (let ((screen (make-array (shape 0 rows 0 cols)))
          (y -1) (x 0))
      (for-each
       (lambda (l)
         (set! y (+ y 1))
         (set! x 0)
         (for-each
          (lambda (c)
            (screen-set! screen y x c)
            (set! x (+ x 1)))
          (string->row l cols)))
       lines)
      screen)))

(provide "term/screen")

