;;;; stty.scm -- stty-like interface to termios

(define-module term.stty
  (export stty with-stty)
  (use gauche.termios))
(select-module term.stty)

(define stty-lookup #f)

(unless stty-lookup
  (set! stty-lookup (make-hash-table 'eq?))
  (for-each
   (lambda (c)
     (let ((type (cadr c))
           (value (caddr c)))
       (hash-table-put! stty-lookup (car c) (cdr c))))

   ;; ripped from the stty man page

   `(;; characters
     (dsusp    char     #f) ; CHAR will send a terminal stop signal once input flushed
     (eof      char     #f) ; CHAR will send an end of file (terminate the input)
     (eol      char     #\newline) ; CHAR will end the line
     (eol2     char     #f) ; alternate CHAR for ending the line
     (erase    char     #\del) ; CHAR will erase the last character typed
     (intr     char     #f) ; CHAR will send an interrupt signal
     (kill     char     #f) ; CHAR will erase the current line
     (lnext    char     #f) ; CHAR will enter the next character quoted
     (quit     char     #f) ; CHAR will send a quit signal
     (rprnt    char     #f) ; CHAR will redraw the current line
     (start    char     #f) ; CHAR will restart the output after stopping it
     (stop     char     #f) ; CHAR will stop the output
     (susp     char     #f) ; CHAR will send a terminal stop signal
     (swtch    char     #f) ; CHAR will switch to a different shell layer
     (werase   char     #f) ; CHAR will erase the last word typed

     ;; special settings
     (cols     special  #f) ; tell the kernel that the terminal has N columns
     (columns  special  #f) ; same as cols N
     (ispeed   special  #f) ; set the input speed to N
     (line     special  #f) ; use line discipline N
     (min      special  #f) ; with -icanon, set N characters minimum for a completed read
     (ospeed   special  #f) ; set the output speed to N
     (rows     special  #f) ; tell the kernel that the terminal has N rows
     (size     special  #f) ; print the number of rows and columns according to the kernel
     (speed    special  #f) ; print the terminal speed
     (time     special  #f) ; with -icanon, set read timeout of N tenths of a second

     ;; control settings
     (clocal   control  ,CLOCAL)  ; disable modem control signals
     (cread    control  ,CREAD)   ; allow input to be received
     (crtscts  control  ,CRTSCTS) ; enable RTS/CTS handshaking
     (cs5      control  ,CS5)     ; set character size to 5 bits
     (cs6      control  ,CS6)     ; set character size to 6 bits
     (cs7      control  ,CS7)     ; set character size to 7 bits
     (cs8      control  ,CS8)     ; set character size to 8 bits
     (cstopb   control  ,CSTOPB)  ; use two stop bits per character (one with `-')
     (hup      control  ,HUPCL)   ; send a hangup signal when the last process closes the tty
     (hupcl    control  ,HUPCL)   ; same as [-]hup
     (parenb   control  ,PARENB)  ; generate parity bit in output and expect parity bit in input
     (parodd   control  ,PARODD)  ; set odd parity (even with `-')

     ;; input settings
     (brkint   input    ,BRKINT)  ; breaks cause an interrupt signal
     (icrnl    input    ,ICRNL)   ; translate carriage return to newline
     (ignbrk   input    ,IGNBRK)  ; ignore break characters
     (igncr    input    ,IGNCR)   ; ignore carriage return
     (ignpar   input    ,IGNPAR)  ; ignore characters with parity errors
     (imaxbel  input    ,IMAXBEL) ; * beep and do not flush a full input buffer on a character
     ;;(inclr    input    ,INCLR)   ; translate newline to carriage return
     (inpck    input    ,INPCK)   ; enable input parity checking
     (istrip   input    ,ISTRIP)  ; clear high (8th) bit of input characters
     (iuclc    input    ,IUCLC)   ; * translate uppercase characters to lowercase
     (ixany    input    ,IXANY)   ; * let any character restart output, not only start character
     (ixoff    input    ,IXOFF)   ; enable sending of start/stop characters
     (ixon     input    ,IXON)    ; enable XON/XOFF flow control
     (parmrk   input    ,PARMRK)  ; mark parity errors (with a 255-0-character sequence)
     (tandem   input    ,IXOFF)   ; same as [-]ixoff

     ;; output settings
     (bs0      output   ,BS0) ; backspace delay style, N in [0..1]
     (bs1      output   ,BS1) ; backspace delay style, N in [0..1]
     (cr0      output   ,CR0) ; carriage return delay style, N in [0..3]
     (cr1      output   ,CR1) ; carriage return delay style, N in [0..3]
     (cr2      output   ,CR2) ; carriage return delay style, N in [0..3]
     (cr3      output   ,CR3) ; carriage return delay style, N in [0..3]
     (ff0      output   ,FF0) ; form feed delay style, N in [0..1]
     (ff1      output   ,FF1) ; form feed delay style, N in [0..1]
     (nl0      output   ,NL0) ; newline delay style, N in [0..1]
     (nl1      output   ,NL1) ; newline delay style, N in [0..1]
     (ocrnl    output   ,OCRNL) ; translate carriage return to newline
     (ofdel    output   ,OFDEL) ; use delete characters for fill instead of null characters
     (ofill    output   ,OFILL) ; use fill (padding) characters instead of timing for delays
     (olcuc    output   ,OLCUC) ; translate lowercase characters to uppercase
     (onlcr    output   ,ONLCR) ; translate newline to carriage return-newline
     (onlret   output   ,ONLRET) ; newline performs a carriage return
     (onocr    output   ,ONOCR) ; do not print carriage returns in the first column
     (opost    output   ,OPOST) ; postprocess output
     (tab0     output   #f) ; horizontal tab delay style, N in [0..3]
     (tab1     output   #f) ; horizontal tab delay style, N in [0..3]
     (tab2     output   #f) ; horizontal tab delay style, N in [0..3]
     (tab3     output   #f) ; horizontal tab delay style, N in [0..3]
     (tabs     output   #f) ; same as tab0
     ;;(-tabs    output   #f) ; same as tab3
     (vt0      output   ,VT0) ; vertical tab delay style, N in [0..1]
     (vt1      output   ,VT1) ; vertical tab delay style, N in [0..1]

     ;; local settings
     (crterase local    ,ECHOE)   ; echo erase characters as backspace-space-backspace
     (crtkill  local    ,ECHOKE)  ; kill all line by obeying the echoprt and echoe settings
     ;;(-crtkill local    #f) ; kill all line by obeying the echoctl and echok settings
     (ctlecho  local    ,ECHOCTL) ; echo control characters in hat notation (`^c')
     (echo     local    ,ECHO)    ; echo input characters
     (echoctl  local    ,ECHOCTL) ; same as [-]ctlecho
     (echoe    local    ,ECHOE)   ; same as [-]crterase
     (echok    local    ,ECHOK)   ; echo a newline after a kill character
     (echoke   local    ,ECHOKE)  ; same as [-]crtkill
     (echonl   local    ,ECHONL)  ; echo newline even if not echoing other characters
     (echoprt  local    ,ECHOPRT) ; echo erased characters backward, between `\' and '/'
     (icanon   local    ,ICANON)  ; enable erase, kill, werase, and rprnt special characters
     ;;(iexten   local    ,IEXTEN)  ; enable non-POSIX special characters
     (isig     local    ,ISIG)    ; enable interrupt, quit, and suspend special characters
     (noflsh   local    ,NOFLSH)  ; disable flushing after interrupt and quit special characters
     (prterase local    ,ECHOPRT) ; same as [-]echoprt
     (tostop   local    ,TOSTOP)  ; stop background jobs that try to write to the terminal
     (xcase    local    ,XCASE)   ; with icanon, escape with `\' for uppercase characters

     ;; combination settings
     (LCASE    combine  '(lcase))
     (cbreak   combine  '(not icanon))
     (cooked   combine  '(brkint ignpar istrip icrnl ixon opost isig icanon))
                                        ; also eof and eol characters
                                        ; to their default values
     (crt      combine  '(echoe echoctl echoke))
     (dec      combine  '(echoe echoctl echoke (not ixany)))
                                        ; also intr ^c erase 0177 kill ^u
     (decctlq  combine  '(ixany))
     (ek       combine  '()) ; erase and kill characters to their default values
     (evenp    combine  '(parenb (not parodd) cs7))
     ;;(-evenp combine  #f) ; same as -parenb cs8
     (lcase    combine  '(xcase iuclc olcuc))
     (litout   combine  '(cs8 (not parenb istrip opost)))
     ;;(-litout  combine  #f) ; same as parenb istrip opost cs7
     (nl       combine  '(not icrnl onlcr))
     ;;(-nl      combine  #f) ; same as icrnl -inlcr -igncr onlcr -ocrnl -onlret
     (oddp     combine  '(parenb parodd cs7))
     (parity   combine  '(evenp)) ; same as [-]evenp
     (pass8    combine  '(cs8 (not parenb istrip)))
     ;;(-pass8   combine  #f) ; same as parenb istrip cs7
     (raw      combine  '(not ignbrk brkint ignpar parmrk
                              inpck istrip inlcr igncr icrnl))
     (ixon     combine  '(ixoff iuclc ixany imaxbel opost isig
                                icanon xcase ))  ; min 1
     ;;(time     combine  #f) ; 0
     ;;(-raw     combine  #f) ; same as cooked
     (sane     combine  '(cread brkint icrnl imaxbel opost onlcr
                                nl0 cr0 bs0 vt0 ff0 isig icanon ; tab0
                                iexten echo echoe echok echoctl echoke
                                (not ignbrk inlcr igncr ixoff iuclc ixany  
                                     olcuc ocrnl onocr onlret ofill ofdel
                                     echonl noflsh xcase tostop echoprt)))
                                        ; plus all special characters to
                                        ; their default values

     )))

(define (stty . args)

  (let* ((port (current-input-port))
         (attr (sys-tcgetattr port))
         (iflag (slot-ref attr 'iflag))
         (oflag (slot-ref attr 'oflag))
         (cflag (slot-ref attr 'cflag))
         (lflag (slot-ref attr 'lflag))
         )

    ;; parse change requests
    (let loop ((lst args)
               (flag #t))
      ;;(format (current-error-port) "loop ~S ~S\n" lst flag)
      (if (null? lst)
        #f
        (let ((command (car lst)))
          (cond
            ((pair? command)    ;; recurse on sub-expr
             (loop command flag))
            ((eq? command 'not) ;; toggle current setting
             (set! flag (not flag)))
            (else
             (let* ((what (hash-table-get stty-lookup command #f))
                    (type (if (pair? what) (car what) #f)))
               (case type
                 ((input)
                  ;;(format (current-error-port) "input ~S: ~S\n" command (cadr what))
                  (if flag
                    (set! iflag (logior iflag (cadr what)))
                    (set! iflag (logand iflag (lognot (cadr what))))))
                 ((output)
                  (if flag
                    (set! oflag (logior oflag (cadr what)))
                    (set! iflag (logand oflag (lognot (cadr what))))))
                 ((control)
                  (if flag
                    (set! cflag (logior cflag (cadr what)))
                    (set! cflag (logand cflag (lognot (cadr what))))))
                 ((local)
                  (if flag
                    (set! lflag (logior lflag (cadr what)))
                    (set! lflag (logand lflag (lognot (cadr what))))))
                 ((special)
                  (error "special settings not yet supported"))
                 ((character)
                  (error "character settings not yet supported"))
                 ((combine)      ;; recurse on def of this command
                  (loop (cadr what) flag))
                 (else
                  #f)))))
          (loop (cdr lst) flag))))

    ;; set new values
    (slot-set! attr 'iflag iflag)
    (slot-set! attr 'oflag oflag)
    (slot-set! attr 'cflag cflag)
    (slot-set! attr 'lflag lflag)
    (sys-tcsetattr port TCSANOW attr)))

(define (with-stty setting thunk)
  (let* ((port (current-input-port))
         (attr (sys-tcgetattr port)))
    (dynamic-wind
        (lambda () (stty setting))
        thunk
        (lambda () (sys-tcsetattr port TCSANOW attr)))))

(provide "term/stty")

